from abc import ABC, abstractmethod
from base64 import b64encode
from copy import deepcopy
from typing import Any, TypeVar, cast

from sdkite.http.adapter import HTTPAdapter, HTTPAdapterSpec
from sdkite.http.model import HTTPRequest

# until mypy has support for Self type
# see https://github.com/python/mypy/issues/11871
_AuthSelf = TypeVar("_AuthSelf", bound="Auth")


class Auth(ABC):
    name: str

    def __init__(
        self,
        http_adapter_spec: HTTPAdapterSpec,
        *,
        interceptor_order: int = 0,
    ) -> None:
        self._http_adapter_spec = http_adapter_spec
        self._interceptor_order = interceptor_order

    def __set_name__(self, _: Any, name: str) -> None:
        self.name = name
        self._http_adapter_spec.register_interceptor(
            "request_interceptor", self.name, self._interceptor_order
        )

    def __get__(self: _AuthSelf, instance: Any, __: Any) -> _AuthSelf:
        if instance is None:
            return self
        attr_name = f"_sdkite_auth_{self.name}"
        try:
            instance_auth = getattr(instance, attr_name)
        except AttributeError:
            instance_auth = deepcopy(self)
            setattr(instance, attr_name, instance_auth)
        return cast(_AuthSelf, instance_auth)

    @abstractmethod
    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        ...


class BasicAuth(Auth):
    def __init__(
        self,
        http_adapter_spec: HTTPAdapterSpec,
        username: str = "",
        password: str = "",
        *,
        interceptor_order: int = 0,
    ) -> None:
        super().__init__(http_adapter_spec, interceptor_order=interceptor_order)
        self.username = username
        self.password = password

    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        # pylint: disable=line-too-long

        # UTF-8 encoding seems to be the de-facto standard in browsers
        # see e.g. https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication#character_encoding_of_http_authentication

        # pylint: enable=line-too-long

        credentials = b64encode(
            f"{self.username}:{self.password}".encode("utf8")
        ).decode()
        request.headers["Authorization"] = f"Basic {credentials}"
        return request


class NoAuth(Auth):
    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        return request



from binascii import crc32
from inspect import Parameter, signature
import sys
from typing import TYPE_CHECKING, Any, List, Optional, Union, cast, get_type_hints
from unittest.mock import Mock, call

import pytest

from sdkite import Pagination, paginated

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import Protocol
else:  # pragma: no cover
    from typing import Protocol

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator


if TYPE_CHECKING:
    from typing_extensions import assert_type
else:
    # no need to have typing_extensions installed
    def assert_type(val: Any, _: Any) -> Any:
        return val


def test_paginated_page() -> None:
    mock_range = Mock(side_effect=range)

    @paginated()
    def fct(pagination: Pagination) -> Iterable[int]:
        page_size = 20
        start = pagination.page * page_size
        stop = start + page_size
        if stop == 100:
            pagination.finish()
        return cast(range, mock_range(start, stop))

    assert list(fct()) == list(range(100))
    assert mock_range.call_args_list == [
        call(0, 20),
        call(20, 40),
        call(40, 60),
        call(60, 80),
        call(80, 100),
    ]


def test_paginated_offset() -> None:
    mock_range = Mock(side_effect=range)

    @paginated()
    def fct(pagination: Pagination) -> Iterable[int]:
        start = pagination.offset
        stop = start * 2 + 1
        if stop > 100:
            pagination.finish()
        return cast(range, mock_range(start, stop))

    assert list(fct()) == list(range(127))
    assert mock_range.call_args_list == [
        call(0, 1),
        call(1, 3),
        call(3, 7),
        call(7, 15),
        call(15, 31),
        call(31, 63),
        call(63, 127),
    ]


def test_paginated_offset_during_iteration() -> None:
    @paginated()
    def fct(pagination: Pagination) -> Iterable[int]:
        for _ in range(10):
            print(pagination.offset)
            yield pagination.offset
            if pagination.offset == 42:
                pagination.finish()
                break

    assert list(fct()) == list(range(42))


def test_paginated_context() -> None:
    @paginated(context=b"00000000")
    def fct(pagination: Pagination) -> Iterable[str]:
        pagination.context = crc32(pagination.context).to_bytes(4, "big")
        value = pagination.context.hex()
        yield value
        if value[0] == "0":
            pagination.finish()

    assert list(fct()) == [
        "c0088d03",
        "4e3ed71e",
        "400c5633",
        "e07be581",
        "1629ad8a",
        "756e02aa",
        "3bd941ec",
        "c4e87e14",
        "d4fca750",
        "be8100bc",
        "fec968a5",
        "d21ba58f",
        "131d3482",
        "cf8a06b5",
        "0e995f9f",
    ]


@pytest.mark.parametrize("stop_when_empty", (None, True, False))
def test_paginated_stop_when_empty(stop_when_empty: Optional[bool]) -> None:
    mock_range = Mock(side_effect=range)
    sizes = [30, 10, 20, 0, 40, 50]

    if stop_when_empty is None:
        decorator = paginated()
    else:
        decorator = paginated(stop_when_empty=stop_when_empty)

    @decorator
    def fct(pagination: Pagination) -> Iterable[int]:
        start = pagination.offset
        try:
            stop = start + sizes[pagination.page]
        except IndexError:
            pagination.finish()
            return []
        return cast(range, mock_range(start, stop))

    # contrary to 'paginated', 'decorator' is not in pylint's 'signature-mutators'
    # pylint: disable-next=no-value-for-parameter
    return_value = fct()

    if stop_when_empty is False:
        assert list(return_value) == list(range(150))
        assert mock_range.call_args_list == [
            call(0, 30),
            call(30, 40),
            call(40, 60),
            call(60, 60),
            call(60, 100),
            call(100, 150),
        ]
    else:
        assert list(return_value) == list(range(60))
        assert mock_range.call_args_list == [
            call(0, 30),
            call(30, 40),
            call(40, 60),
            call(60, 60),
        ]


def test_paginated_wrapping_function() -> None:
    @paginated()
    def fct(
        pagination: Pagination,
        var_a: int,
        *var_b: int,
        var_c: int,
        var_d: int = 42,
        **var_e: int,
    ) -> List[int]:
        """Foobar"""
        if pagination.page == 0:
            return [var_a, *var_b, var_c, var_d, *var_e.values()]
        return []

    class ExpectedType(Protocol):
        def __call__(
            self,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> Iterator[int]:
            ...

    assert_type(fct, ExpectedType)

    assert fct.__name__ == "fct", "name"
    assert fct.__doc__ == "Foobar", "doc"
    sig = signature(fct)
    assert list(sig.parameters.values()) == [
        Parameter("var_a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        Parameter("var_b", Parameter.VAR_POSITIONAL, annotation=int),
        Parameter("var_c", Parameter.KEYWORD_ONLY, annotation=int),
        Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int, default=42),
        Parameter("var_e", Parameter.VAR_KEYWORD, annotation=int),
    ], "signature parameters"
    assert sig.return_annotation == Iterator[Any], "signature return_annotation"
    assert get_type_hints(fct) == {
        "var_a": int,
        "var_b": int,
        "var_c": int,
        "var_d": int,
        "var_e": int,
        "return": Iterator[Any],
    }

    assert list(fct(1, 2, 3, foo=4, bar=5, var_c=6)) == [1, 2, 3, 6, 42, 4, 5]


def test_paginated_wrapping_method() -> None:
    class Klass:
        def __init__(self, value: int) -> None:
            self.value = value

        @paginated()
        def meth(
            self,
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [self.value, var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

        # assert_type(Klass.meth, ...) performed other file test due to PEP 570 requirement

    assert Klass.meth.__name__ == "meth", "name"
    assert Klass.meth.__doc__ == "Foobar", "doc"
    sig = signature(Klass.meth)
    assert list(sig.parameters.values()) == [
        Parameter("self", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("var_a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        Parameter("var_b", Parameter.VAR_POSITIONAL, annotation=int),
        Parameter("var_c", Parameter.KEYWORD_ONLY, annotation=int),
        Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int, default=42),
        Parameter("var_e", Parameter.VAR_KEYWORD, annotation=int),
    ], "signature parameters"
    assert sig.return_annotation == Iterator[Any], "signature return_annotation"
    assert get_type_hints(Klass.meth) == {
        "var_a": int,
        "var_b": int,
        "var_c": int,
        "var_d": int,
        "var_e": int,
        "return": Iterator[Any],
    }

    instance = Klass(1337)

    class ExpectedType(Protocol):
        def __call__(
            self,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> Iterator[int]:
            ...

    assert_type(instance.meth, ExpectedType)

    assert instance.meth.__name__ == "meth", "name"
    assert instance.meth.__doc__ == "Foobar", "doc"
    sig = signature(instance.meth)
    assert list(sig.parameters.values()) == [
        Parameter("var_a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        Parameter("var_b", Parameter.VAR_POSITIONAL, annotation=int),
        Parameter("var_c", Parameter.KEYWORD_ONLY, annotation=int),
        Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int, default=42),
        Parameter("var_e", Parameter.VAR_KEYWORD, annotation=int),
    ], "signature parameters"
    assert sig.return_annotation == Iterator[Any], "signature return_annotation"
    assert get_type_hints(instance.meth) == {
        "var_a": int,
        "var_b": int,
        "var_c": int,
        "var_d": int,
        "var_e": int,
        "return": Iterator[Any],
    }

    assert list(instance.meth(1, 2, 3, foo=4, bar=5, var_c=6)) == [
        1337,
        1,
        2,
        3,
        6,
        42,
        4,
        5,
    ]


def test_paginated_wrapping_staticmethod() -> None:
    class Klass:
        @staticmethod
        @paginated()
        def meth0(
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

        @paginated()
        @staticmethod
        def meth1(
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

    class ExpectedType(Protocol):
        def __call__(
            self,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> Iterator[int]:
            ...

    instance = Klass()

    assert_type(Klass.meth0, ExpectedType)
    assert_type(Klass.meth1, ExpectedType)
    assert_type(instance.meth0, ExpectedType)
    assert_type(instance.meth1, ExpectedType)

    assert Klass.meth0.__name__ == "meth0", "name"
    assert Klass.meth1.__name__ == "meth1", "name"
    assert instance.meth0.__name__ == "meth0", "name"
    assert instance.meth1.__name__ == "meth1", "name"

    for meth in (Klass.meth0, Klass.meth1, instance.meth0, instance.meth1):
        assert meth.__doc__ == "Foobar", "doc"
        sig = signature(meth)
        assert list(sig.parameters.values()) == [
            Parameter("var_a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
            Parameter("var_b", Parameter.VAR_POSITIONAL, annotation=int),
            Parameter("var_c", Parameter.KEYWORD_ONLY, annotation=int),
            Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int, default=42),
            Parameter("var_e", Parameter.VAR_KEYWORD, annotation=int),
        ], "signature parameters"
        assert sig.return_annotation == Iterator[Any], "signature return_annotation"
        assert get_type_hints(meth) == {
            "var_a": int,
            "var_b": int,
            "var_c": int,
            "var_d": int,
            "var_e": int,
            "return": Iterator[Any],
        }

        assert list(meth(1, 2, 3, foo=4, bar=5, var_c=6)) == [1, 2, 3, 6, 42, 4, 5]


def test_paginated_wrapping_classmethod() -> None:
    class Klass:
        value = 1337

        @classmethod
        @paginated()
        def meth0(
            cls,
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [cls.value, var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

        @paginated()
        @classmethod
        def meth1(
            cls,
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [cls.value, var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

    class ExpectedType(Protocol):
        def __call__(
            self,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> Iterator[int]:
            ...

    assert_type(Klass.meth0, ExpectedType)
    assert_type(Klass.meth1, ExpectedType)

    assert Klass.meth0.__name__ == "meth0", "name"
    assert Klass.meth1.__name__ == "meth1", "name"

    for meth in (Klass.meth0, Klass.meth1):
        assert meth.__doc__ == "Foobar", "doc"
        sig = signature(meth)
        assert list(sig.parameters.values()) == [
            Parameter("var_a", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
            Parameter("var_b", Parameter.VAR_POSITIONAL, annotation=int),
            Parameter("var_c", Parameter.KEYWORD_ONLY, annotation=int),
            Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int, default=42),
            Parameter("var_e", Parameter.VAR_KEYWORD, annotation=int),
        ], "signature parameters"
        assert sig.return_annotation == Iterator[Any], "signature return_annotation"
        assert get_type_hints(meth) == {
            "var_a": int,
            "var_b": int,
            "var_c": int,
            "var_d": int,
            "var_e": int,
            "return": Iterator[Any],
        }

        assert list(meth(1, 2, 3, foo=4, bar=5, var_c=6)) == [
            1337,
            1,
            2,
            3,
            6,
            42,
            4,
            5,
        ]


def test_paginated_not_iterable() -> None:
    @paginated()  # type: ignore[arg-type]
    def fct(pagination: Pagination) -> Union[Iterable[int], int]:
        if pagination.page < 10:
            return [pagination.page]
        return 42

    output = fct()

    with pytest.raises(TypeError) as excinfo:
        list(output)

    assert str(excinfo.value) == "'int' object is not iterable"


def test_paginated_no_param() -> None:
    with pytest.raises(TypeError) as excinfo:

        @paginated()  # type: ignore[arg-type]
        def fct() -> Iterable[int]:
            yield 42

    assert (
        str(excinfo.value)
        == "Paginated function 'fct' must have 'pagination' as first parameter"
    )


def test_paginated_wrong_param_name() -> None:
    with pytest.raises(TypeError) as excinfo:

        @paginated()
        def fct(
            # pylint: disable=unused-argument
            param0: Pagination,
            param1: int,
        ) -> Iterable[int]:
            yield 42

    assert (
        str(excinfo.value)
        == "Paginated function 'fct' must have 'pagination' as first parameter"
    )


def test_paginated_wrong_param_kind() -> None:
    with pytest.raises(TypeError) as excinfo:

        @paginated()  # type: ignore[arg-type]
        def fct(
            # pylint: disable=unused-argument
            *,
            pagination: Pagination,
        ) -> Iterable[int]:
            yield 42

    assert (
        str(excinfo.value)
        == "Paginated function 'fct' must have 'pagination' as first parameter"
    )


def test_paginated_called_with_pagination_param() -> None:
    @paginated()
    def fct(
        # pylint: disable=unused-argument
        pagination: Pagination,
        *args: int,
        **kwargs: int,
    ) -> Iterable[int]:
        if pagination.page == 0:
            yield 42

    assert list(fct(13, param=37)) == [42]

    with pytest.raises(TypeError) as excinfo:
        list(fct(13, pagination=37))

    assert (
        str(excinfo.value)
        == "Paginated function 'fct' cannot be called with a 'pagination' parameter"
    )


def test_paginated_no_type_hint() -> None:
    @paginated()
    def fct(pagination):  # type: ignore
        if pagination.page == 4:
            pagination.finish()
        return range(pagination.offset, pagination.offset + 10)

    assert list(fct()) == list(range(50))



from os import chdir, getcwd
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple, cast

import pytest
from requests_mock import Mocker

from sdkite import Client


@pytest.fixture(scope="module", autouse=True)
def import_pytest(doctest_namespace: Dict[str, object]) -> None:
    doctest_namespace["Iterator"] = Iterator
    doctest_namespace["Path"] = Path

    doctest_namespace["Client"] = Client

    def list_spells(
        max_price: int,
        offset: int = 0,
        page: Optional[int] = None,
    ) -> List[str]:
        assert max_price == 50
        limit = 3
        if page is not None:
            offset = page * limit
        return [
            "Crushing Burden Touch",
            "Great Burden of Sin",
            "Heavy Burden",
            "Strong Feather",
            "Tinur's Hoptoad",
            "Ulms's Juicedaw's Feather",
            "Far Silence",
            "Soul Trap",
        ][offset:][:limit]

    def list_spells_with_ref(
        max_price: Optional[int] = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[str], Optional[str]]:
        assert max_price == 50
        cursors = [None, "57656c636", "f6d652074", "6f2042616", "c6d6f7261"]
        index = cursors.index(cursor)
        return list_spells(max_price, page=index), cursors[index + 1]

    doctest_namespace["list_spells"] = list_spells
    doctest_namespace["list_spells_with_ref"] = list_spells_with_ref


@pytest.fixture(autouse=True)
def patch_http(requests_mock: Mocker) -> None:
    #
    # quickstart
    #
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/world/npc/interact?name=ranis",
        text="Have you found the Telvanni spy?",
    )
    requests_mock.register_uri(
        "POST",
        "https://api.example.com/world/npc/search",
        additional_matcher=lambda request: cast(
            bool,
            request.json() == {"city": "Balmora", "faction": "Mage", "min_level": 8},
        ),
        json=["marayn", "masalinie", "ranis"],
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/world/book/content?id=bk_words_of_the_wind",
        request_headers={"Authorization": "Basic TmVyZXZhcmluZTpJbmNhcm5hdGU="},
        text="Words of the Wind\n\nA volume of verse collected from...",
    )

    #
    # http_auth
    #
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/whoami",
        request_headers={"Authorization": "Basic QWxpY2U6VzBuZGVybEBuZA=="},
        text="Welcome Alice!",
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/noauth",
        additional_matcher=lambda request: "Authorization" not in request.headers,
        text="The /noauth endpoint has been called without auth",
    )


@pytest.fixture(autouse=True)
def change_directory(tmp_path: Path) -> Iterator[None]:
    cwd = getcwd()
    try:
        chdir(tmp_path)
        yield
    finally:
        chdir(cwd)



from pathlib import Path
import sys
from typing import List

import pytest


def pytest_collection_modifyitems(
    # pylint: disable=unused-argument
    config: pytest.Config,
    items: List[pytest.Item],
) -> None:
    root = Path(__file__).parent.parent
    for item in items:
        relative = Path(item.fspath).parent.relative_to(root)
        mark = next(part for part in relative.parts if part != "tests")
        item.add_marker(getattr(pytest.mark, mark))


# skipping some files due to syntax not yet supported
collect_ignore: List[str] = []
if sys.version_info < (3, 8):
    collect_ignore.append("unit/test_pagination_pep570.py")



from dataclasses import dataclass
import secrets
from typing import Dict, List, Optional, Tuple, Union

import pytest

from sdkite.http import HTTPBodyEncoding
from sdkite.http.utils import encode_request_body


@pytest.fixture(autouse=True)
def patched_adapter(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(secrets, "token_hex", lambda size: "0" * size)


@dataclass
class Ok:
    data: Optional[bytes]
    content_type: Optional[str] = None
    is_iterator: bool = False


@dataclass
class Ko:
    err_type: str
    err_path: Optional[str] = None

    def msg(self, encoding: HTTPBodyEncoding) -> str:
        error_msg = "Cannot encode body"
        if self.err_path:
            error_msg += f" ({self.err_path})"
        error_msg += f" of type '{self.err_type}'"
        error_msg += f" with '{encoding.name}' encoding"
        return error_msg


class CustomClass:
    pass


CUSTOM_CLASS_INSTANCE = CustomClass()


# the type annotation makes sure deepest values are Ok or Ko instances
_CASES_DESCRIPTIONS: List[Tuple[str, object, Dict[HTTPBodyEncoding, Union[Ok, Ko]]]] = [
    (
        "None",
        None,
        {
            HTTPBodyEncoding.AUTO: Ok(b""),
            HTTPBodyEncoding.NONE: Ok(b""),
            HTTPBodyEncoding.JSON: Ok(b"null", "application/json"),
            HTTPBodyEncoding.URLENCODE: Ok(
                b"",
                "application/x-www-form-urlencoded",
            ),
            HTTPBodyEncoding.MULTIPART: Ko("NoneType"),
        },
    ),
    (
        "True",
        True,
        {
            HTTPBodyEncoding.AUTO: Ok(b"true", "application/json"),
            HTTPBodyEncoding.NONE: Ko("bool"),
            HTTPBodyEncoding.JSON: Ok(b"true", "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("bool"),
            HTTPBodyEncoding.MULTIPART: Ko("bool"),
        },
    ),
    (
        "False",
        False,
        {
            HTTPBodyEncoding.AUTO: Ok(b"false", "application/json"),
            HTTPBodyEncoding.NONE: Ko("bool"),
            HTTPBodyEncoding.JSON: Ok(b"false", "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("bool"),
            HTTPBodyEncoding.MULTIPART: Ko("bool"),
        },
    ),
    (
        "int",
        42,
        {
            HTTPBodyEncoding.AUTO: Ok(b"42", "application/json"),
            HTTPBodyEncoding.NONE: Ko("int"),
            HTTPBodyEncoding.JSON: Ok(b"42", "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("int"),
            HTTPBodyEncoding.MULTIPART: Ko("int"),
        },
    ),
    (
        "float",
        13.37,
        {
            HTTPBodyEncoding.AUTO: Ok(b"13.37", "application/json"),
            HTTPBodyEncoding.NONE: Ko("float"),
            HTTPBodyEncoding.JSON: Ok(b"13.37", "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("float"),
            HTTPBodyEncoding.MULTIPART: Ko("float"),
        },
    ),
    (
        "bytes",
        b"\xc3\xa6ther[1+2]%3",
        {
            HTTPBodyEncoding.AUTO: Ok(b"\xc3\xa6ther[1+2]%3"),
            HTTPBodyEncoding.NONE: Ok(b"\xc3\xa6ther[1+2]%3"),
            HTTPBodyEncoding.JSON: Ko("bytes"),
            HTTPBodyEncoding.URLENCODE: Ko("bytes"),
            HTTPBodyEncoding.MULTIPART: Ko("bytes"),
        },
    ),
    (
        "str",
        "æther[1+2]%3",
        {
            HTTPBodyEncoding.AUTO: Ok(b"\xc3\xa6ther[1+2]%3"),
            HTTPBodyEncoding.NONE: Ok(b"\xc3\xa6ther[1+2]%3"),
            HTTPBodyEncoding.JSON: Ok(b'"\\u00e6ther[1+2]%3"', "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("str"),
            HTTPBodyEncoding.MULTIPART: Ko("str"),
        },
    ),
    (
        "list[misc]",
        [42, True, b"foo", "bar"],
        {
            HTTPBodyEncoding.AUTO: Ko("bytes", "2"),
            HTTPBodyEncoding.NONE: Ko("list"),
            HTTPBodyEncoding.JSON: Ko("bytes", "2"),
            HTTPBodyEncoding.URLENCODE: Ko("list"),
            HTTPBodyEncoding.MULTIPART: Ko("list"),
        },
    ),
    (
        "list[str]",
        ["æther", "foo"],
        {
            HTTPBodyEncoding.AUTO: Ok(
                b'["\\u00e6ther","foo"]',
                "application/json",
            ),
            HTTPBodyEncoding.NONE: Ko("list"),
            HTTPBodyEncoding.JSON: Ok(
                b'["\\u00e6ther","foo"]',
                "application/json",
            ),
            HTTPBodyEncoding.URLENCODE: Ko("list"),
            HTTPBodyEncoding.MULTIPART: Ko("list"),
        },
    ),
    (
        "tuple[str]",
        ("æther", "foo"),
        {
            HTTPBodyEncoding.AUTO: Ok(
                b'["\\u00e6ther","foo"]',
                "application/json",
            ),
            HTTPBodyEncoding.NONE: Ko("tuple"),
            HTTPBodyEncoding.JSON: Ok(
                b'["\\u00e6ther","foo"]',
                "application/json",
            ),
            HTTPBodyEncoding.URLENCODE: Ko("tuple"),
            HTTPBodyEncoding.MULTIPART: Ko("tuple"),
        },
    ),
    (
        "set[str]",
        {"æther", "foo"},
        {
            HTTPBodyEncoding.AUTO: Ok(
                b'["foo","\\u00e6ther"]',
                "application/json",
            ),
            HTTPBodyEncoding.NONE: Ko("set"),
            HTTPBodyEncoding.JSON: Ok(
                b'["foo","\\u00e6ther"]',
                "application/json",
            ),
            HTTPBodyEncoding.URLENCODE: Ko("set"),
            HTTPBodyEncoding.MULTIPART: Ko("set"),
        },
    ),
    (
        "dict[bytes,bytes]",
        {b"\xc3\xa6ther": b"foobar", b"a[b=c]+%3": b"d[e=f]%+4"},
        {
            HTTPBodyEncoding.AUTO: Ok(
                b"a%5Bb%3Dc%5D%2B%253=d%5Be%3Df%5D%25%2B4&%C3%A6ther=foobar",
                "application/x-www-form-urlencoded",
            ),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ko("bytes", "b'a[b=c]+%3'>:key"),
            HTTPBodyEncoding.URLENCODE: Ok(
                b"a%5Bb%3Dc%5D%2B%253=d%5Be%3Df%5D%25%2B4&%C3%A6ther=foobar",
                "application/x-www-form-urlencoded",
            ),
            HTTPBodyEncoding.MULTIPART: Ok(
                (
                    b"------00000000000000000000000000000000\r\n"
                    b'Content-Disposition: form-data; name="a%5Bb%3Dc%5D%2B%253"\r\n'
                    b"Content-Type: application/octet-stream\r\n"
                    b"\r\n"
                    b"d[e=f]%+4"
                    b"\r\n"
                    b"------00000000000000000000000000000000\r\n"
                    b'Content-Disposition: form-data; name="%C3%A6ther"\r\n'
                    b"Content-Type: application/octet-stream\r\n"
                    b"\r\n"
                    b"foobar"
                    b"\r\n"
                    b"------00000000000000000000000000000000--\r\n"
                ),
                "multipart/form-data; boundary=----00000000000000000000000000000000",
            ),
        },
    ),
    (
        "dict[str,str]",
        {"æther": "foobar", "a[b=c]+%3": "d[e=f]%+4"},
        {
            HTTPBodyEncoding.AUTO: Ok(
                b'{"a[b=c]+%3":"d[e=f]%+4","\\u00e6ther":"foobar"}',
                "application/json",
            ),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ok(
                b'{"a[b=c]+%3":"d[e=f]%+4","\\u00e6ther":"foobar"}',
                "application/json",
            ),
            HTTPBodyEncoding.URLENCODE: Ok(
                b"a%5Bb%3Dc%5D%2B%253=d%5Be%3Df%5D%25%2B4&%C3%A6ther=foobar",
                "application/x-www-form-urlencoded",
            ),
            HTTPBodyEncoding.MULTIPART: Ok(
                (
                    b"------00000000000000000000000000000000\r\n"
                    b'Content-Disposition: form-data; name="a%5Bb%3Dc%5D%2B%253"\r\n'
                    b"Content-Type: application/octet-stream\r\n"
                    b"\r\n"
                    b"d[e=f]%+4"
                    b"\r\n"
                    b"------00000000000000000000000000000000\r\n"
                    b'Content-Disposition: form-data; name="%C3%A6ther"\r\n'
                    b"Content-Type: application/octet-stream\r\n"
                    b"\r\n"
                    b"foobar"
                    b"\r\n"
                    b"------00000000000000000000000000000000--\r\n"
                ),
                "multipart/form-data; boundary=----00000000000000000000000000000000",
            ),
        },
    ),
    (
        "dict[str,int]",
        {"uvw": 13, "xyz": 37},
        {
            HTTPBodyEncoding.AUTO: Ok(b'{"uvw":13,"xyz":37}', "application/json"),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ok(b'{"uvw":13,"xyz":37}', "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("int", "'uvw'"),
            HTTPBodyEncoding.MULTIPART: Ko("int", "'uvw'"),
        },
    ),
    (
        "dict[int,str]",
        {13: "37", 42: "42"},
        {
            HTTPBodyEncoding.AUTO: Ok(b'{13:"37",42:"42"}', "application/json"),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ok(b'{13:"37",42:"42"}', "application/json"),
            HTTPBodyEncoding.URLENCODE: Ko("int", "13>:key"),
            HTTPBodyEncoding.MULTIPART: Ko("int", "13>:key"),
        },
    ),
    (
        "dict[str,CustomClass]",
        {"xyz": CUSTOM_CLASS_INSTANCE},
        {
            HTTPBodyEncoding.AUTO: Ko("CustomClass", "'xyz'"),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ko("CustomClass", "'xyz'"),
            HTTPBodyEncoding.URLENCODE: Ko("CustomClass", "'xyz'"),
            HTTPBodyEncoding.MULTIPART: Ko("CustomClass", "'xyz'"),
        },
    ),
    (
        "dict[CustomClass,str]",
        {CUSTOM_CLASS_INSTANCE: "42"},
        {
            HTTPBodyEncoding.AUTO: Ko("CustomClass", f"{CUSTOM_CLASS_INSTANCE}>:key"),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ko("CustomClass", f"{CUSTOM_CLASS_INSTANCE}>:key"),
            HTTPBodyEncoding.URLENCODE: Ko(
                "CustomClass", f"{CUSTOM_CLASS_INSTANCE}>:key"
            ),
            HTTPBodyEncoding.MULTIPART: Ko(
                "CustomClass", f"{CUSTOM_CLASS_INSTANCE}>:key"
            ),
        },
    ),
    (
        "dict[str,dict[str,str]]",
        {"æther": "foobar", "abc": {"uvw": "xyz"}},
        {
            HTTPBodyEncoding.AUTO: Ok(
                b'{"abc":{"uvw":"xyz"},"\\u00e6ther":"foobar"}',
                "application/json",
            ),
            HTTPBodyEncoding.NONE: Ko("dict"),
            HTTPBodyEncoding.JSON: Ok(
                b'{"abc":{"uvw":"xyz"},"\\u00e6ther":"foobar"}',
                "application/json",
            ),
            HTTPBodyEncoding.URLENCODE: Ko("dict", "'abc'"),
            HTTPBodyEncoding.MULTIPART: Ko("dict", "'abc'"),
        },
    ),
    (
        "CustomClass",
        CUSTOM_CLASS_INSTANCE,
        {
            HTTPBodyEncoding.AUTO: Ko("CustomClass"),
            HTTPBodyEncoding.NONE: Ko("CustomClass"),
            HTTPBodyEncoding.JSON: Ko("CustomClass"),
            HTTPBodyEncoding.URLENCODE: Ko("CustomClass"),
            HTTPBodyEncoding.MULTIPART: Ko("CustomClass"),
        },
    ),
]


@pytest.mark.parametrize(
    "body, encoding, expected",
    [
        pytest.param(
            body,
            encoding,
            value,
            id=f"{encoding.name}-{name}",
        )
        for name, body, by_encoding in _CASES_DESCRIPTIONS
        for encoding, value in by_encoding.items()
    ],
)
def test_cases(
    body: object,
    encoding: HTTPBodyEncoding,
    expected: Union[Ok, Ko],
) -> None:
    if isinstance(expected, Ok):
        data, content_type = encode_request_body(body, encoding)
        assert isinstance(data, bytes) != expected.is_iterator
        if not isinstance(data, bytes):
            data = b"".join(data)
        assert data == expected.data, "Invalid data"
        assert content_type == expected.content_type, "Invalid content type"

    else:
        with pytest.raises(TypeError) as excinfo:
            encode_request_body(body, encoding)

        assert excinfo.type is TypeError
        assert str(excinfo.value) == expected.msg(encoding)


def test_invalid_encoding_type() -> None:
    with pytest.raises(TypeError):
        encode_request_body(None, "oops")  # type: ignore[arg-type]



from typing import Optional, get_type_hints


class Client:
    def __init__(
        self,
    ) -> None:
        self._parent: Optional[Client] = None

        # auto-init sub-clients
        cls = type(self)
        for attr_name, attr_type in get_type_hints(cls).items():
            if not hasattr(cls, attr_name) and issubclass(attr_type, Client):
                if attr_type.__init__ is not Client.__init__:
                    raise TypeError(
                        f"Class {attr_type.__name__} used as {cls.__name__}.{attr_name}"
                        " defines a custom __init__"
                    )
                try:
                    client = attr_type()
                except RecursionError as ex:
                    raise TypeError(
                        "Clients refer each other (found for"
                        f" {attr_type.__name__} used as {cls.__name__}.{attr_name})"
                    ) from ex
                client._parent = self
                setattr(self, attr_name, client)



from contextlib import contextmanager
from functools import reduce
import json
import secrets
import sys
from typing import Optional, Tuple, Union
from urllib.parse import quote_plus
from urllib.parse import urljoin as _urljoin

from sdkite.http.model import HTTPBodyEncoding

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator

if sys.version_info < (3, 10):  # pragma: no cover
    NoneType = type(None)
else:  # pragma: no cover
    from types import NoneType


def urljoin(base: Optional[str], url: Optional[str]) -> Optional[str]:
    """
    Re-implementation of urljoin with some edge cases:
     - Handling of None values (note: this was the case with stdlib but undocumented)
     - Work more like os.path.join when base does not end with "/"
       e.g. __urljoin("https://example.com/foo", "bar")
            gives "https://example.com/foo/bar"
            and not "https://example.com/bar"
    """
    if base:
        if url:
            if not base.endswith("/"):
                base += "/"
            url = _urljoin(base, url)
        else:
            url = base
    elif not url:
        url = None
    return url


def urlsjoin(parts: Iterable[Optional[str]]) -> Optional[str]:
    return reduce(urljoin, parts, None)


def urlencode(data: Union[str, bytes]) -> bytes:
    return quote_plus(data).encode()


_EMPTY_ITERATOR = iter(())


class _VisitorTypeError(TypeError):
    def __init__(self, *path: str, obj: object) -> None:
        self.path = path
        self.obj = obj

    def context(self, path: str) -> "_VisitorTypeError":
        return _VisitorTypeError(path, *self.path, obj=self.obj)

    # the path item to use when the error happens in a mapping key
    MAPPING_KEY_PATH = ":key"


@contextmanager
def _visitor_obj(path: str, obj: object) -> Iterator[None]:
    try:
        yield
    except _VisitorTypeError as ex:
        raise ex.context(path)
    except TypeError:
        raise _VisitorTypeError(path, obj=obj)  # pylint: disable=raise-missing-from


class _Visitor:
    def visit(self, obj: object) -> Iterator[bytes]:
        try:
            for klass in (NoneType, str, bytes, bool, int, float):
                if isinstance(obj, klass):
                    yield from self._visit_raw(obj)
                    return
            if isinstance(obj, (list, tuple, set)):
                yield from self._visit_sequence_start()
                first = True
                for i, value in enumerate(sorted(obj) if isinstance(obj, set) else obj):
                    with _visitor_obj(repr(i), obj=value):
                        yield from self._visit_sequence_item_start(first)
                        yield from self.visit(value)
                        yield from self._visit_sequence_item_end(first)
                    first = False
                yield from self._visit_sequence_end()
                return
            if isinstance(obj, dict):
                yield from self._visit_mapping_start()
                first = True
                for key in sorted(obj):
                    with _visitor_obj(repr(key), obj=obj[key]):
                        yield from self._visit_mapping_item_start(key, first)
                        yield from self.visit(obj[key])
                        yield from self._visit_mapping_item_end(key, first)
                    first = False
                yield from self._visit_mapping_end()
                return
            raise TypeError
        except _VisitorTypeError:
            raise
        except TypeError:
            raise _VisitorTypeError(obj=obj)  # pylint: disable=raise-missing-from

    # pylint: disable=unused-argument

    def _visit_raw(self, obj: object) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_sequence_start(self) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_sequence_end(self) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_sequence_item_start(self, first: bool) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_sequence_item_end(self, first: bool) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_mapping_start(self) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_mapping_end(self) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_mapping_item_start(self, key: object, first: bool) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover

    def _visit_mapping_item_end(self, key: object, first: bool) -> Iterator[bytes]:
        return _EMPTY_ITERATOR  # pragma: no cover


class _VisitorNone(_Visitor):
    def _visit_raw(self, obj: object) -> Iterator[bytes]:
        if obj is None:
            return
        if isinstance(obj, bytes):
            yield obj
            return
        if isinstance(obj, str):
            yield obj.encode()
            return
        raise TypeError

    def _visit_sequence_start(self) -> Iterator[bytes]:
        raise TypeError

    def _visit_mapping_start(self) -> Iterator[bytes]:
        raise TypeError


class _VisitorJSON(_Visitor):
    def _visit_raw(self, obj: object) -> Iterator[bytes]:
        yield json.dumps(obj).encode()

    def _visit_sequence_start(self) -> Iterator[bytes]:
        yield b"["

    def _visit_sequence_end(self) -> Iterator[bytes]:
        yield b"]"

    def _visit_sequence_item_start(self, first: bool) -> Iterator[bytes]:
        if not first:
            yield b","

    def _visit_mapping_start(self) -> Iterator[bytes]:
        yield b"{"

    def _visit_mapping_end(self) -> Iterator[bytes]:
        yield b"}"

    def _visit_mapping_item_start(self, key: object, first: bool) -> Iterator[bytes]:
        if not first:
            yield b","
        with _visitor_obj(_VisitorTypeError.MAPPING_KEY_PATH, obj=key):
            yield from self._visit_raw(key)
        yield b":"


class _VisitorURLEncode(_Visitor):
    root = True

    def _visit_raw(self, obj: object) -> Iterator[bytes]:
        if obj is None:
            return
        if self.root:
            raise TypeError
        if isinstance(obj, (bytes, str)):
            yield urlencode(obj)
            return
        raise TypeError

    def _visit_sequence_start(self) -> Iterator[bytes]:
        raise TypeError

    def _visit_mapping_start(self) -> Iterator[bytes]:
        if not self.root:
            raise TypeError
        self.root = False
        return _EMPTY_ITERATOR

    def _visit_mapping_item_start(self, key: object, first: bool) -> Iterator[bytes]:
        if not first:
            yield b"&"
        with _visitor_obj(_VisitorTypeError.MAPPING_KEY_PATH, obj=key):
            yield from self._visit_raw(key)
        yield b"="


class _VisitorMultipart(_Visitor):
    root = True

    def __init__(self) -> None:
        self.boundary = b"----%s" % secrets.token_hex(32).encode()

    def _visit_raw(self, obj: object) -> Iterator[bytes]:
        if self.root:
            raise TypeError
        if isinstance(obj, bytes):
            yield obj
            return
        if isinstance(obj, str):
            yield obj.encode()
            return
        raise TypeError

    def _visit_sequence_start(self) -> Iterator[bytes]:
        raise TypeError

    def _visit_mapping_start(self) -> Iterator[bytes]:
        if not self.root:
            raise TypeError
        self.root = False
        return _EMPTY_ITERATOR

    def _visit_mapping_item_start(self, key: object, first: bool) -> Iterator[bytes]:
        with _visitor_obj(_VisitorTypeError.MAPPING_KEY_PATH, obj=key):
            name = b"".join(self._visit_raw(key))
        yield (
            b"--%s\r\n"
            b'Content-Disposition: form-data; name="%s"\r\n'
            b"Content-Type: application/octet-stream\r\n"
            b"\r\n"
        ) % (
            self.boundary,
            urlencode(name),
        )

    def _visit_mapping_item_end(self, key: object, first: bool) -> Iterator[bytes]:
        yield b"\r\n"

    def _visit_mapping_end(self) -> Iterator[bytes]:
        yield b"--%s--\r\n" % self.boundary


def encode_request_body(
    body: object,
    encoding: HTTPBodyEncoding,
) -> Tuple[Union[bytes, Iterator[bytes]], Optional[str]]:
    content_type: Optional[str] = None
    original_encoding = encoding

    if encoding == HTTPBodyEncoding.AUTO:
        encoding = HTTPBodyEncoding.JSON
        if isinstance(body, (bytes, str, NoneType)):
            encoding = HTTPBodyEncoding.NONE
        elif isinstance(body, dict):
            if any(isinstance(key, bytes) for key in body):
                encoding = HTTPBodyEncoding.URLENCODE

    visitor: _Visitor
    if encoding == HTTPBodyEncoding.NONE:
        visitor = _VisitorNone()
    elif encoding == HTTPBodyEncoding.JSON:
        visitor = _VisitorJSON()
        content_type = "application/json"
    elif encoding == HTTPBodyEncoding.URLENCODE:
        visitor = _VisitorURLEncode()
        content_type = "application/x-www-form-urlencoded"
    elif encoding == HTTPBodyEncoding.MULTIPART:
        visitor = _VisitorMultipart()
        content_type = f"multipart/form-data; boundary={visitor.boundary.decode()}"
    else:
        raise TypeError("Invalid encoding type")

    try:
        iterator = visitor.visit(body)
        data = b"".join(iterator)
    except _VisitorTypeError as ex:
        error_msg = "Cannot encode body"
        if ex.path:
            error_msg += f" ({'>'.join(ex.path)})"
        error_msg += f" of type '{type(ex.obj).__name__}'"
        error_msg += f" with '{original_encoding.name}' encoding"
        raise TypeError(error_msg)  # pylint: disable=raise-missing-from

    return data, content_type



import sys
from typing import List, cast

from requests_mock import Mocker

from sdkite import Client, Pagination, paginated
from sdkite.http import BasicAuth, HTTPAdapterSpec, NoAuth

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class ApiPublic(Client):
    _http = HTTPAdapterSpec(url="public")
    _auth = NoAuth(_http)

    def version(self) -> str:
        response = self._http.get("version")
        return response.data_str


class User(TypedDict):
    name: str
    age: int


class ApiUsers(Client):
    _http = HTTPAdapterSpec(url="users", headers={"X-Toto": "Abc"})

    def get(self, user_id: int) -> User:
        response = self._http.get(str(user_id))
        return cast(User, response.data_json)

    @paginated(page=1)
    def get_all(self, pagination: Pagination) -> List[User]:
        response = self._http.get(f"all/{pagination.page}")
        return cast(List[User], response.data_json)


class Api(Client):
    _http = HTTPAdapterSpec()
    _auth = BasicAuth(_http, username="tests")

    public: ApiPublic
    users: ApiUsers

    def __init__(self, url: str, password: str) -> None:
        super().__init__()
        self._http.url = url
        self._auth.password = password


def test_http(requests_mock: Mocker) -> None:
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/public/version",
        additional_matcher=lambda request: "Authorization" not in request.headers,
        content=b"1.2.3",
    )

    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/1337",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'{"name":"John Doe","age":42}',
    )

    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/1",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'[{"name":"John Doe","age":42},{"name":"Alice Doe","age":41}]',
    )
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/2",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'[{"name":"Bob Doe","age":10},{"name":"Carole Doe","age":12}]',
    )
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/3",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b"[]",
    )

    client = Api("https://www.example.com/api/v1", "s3cr3t")

    assert client.public.version() == "1.2.3"

    user = client.users.get(1337)
    assert user == User(name="John Doe", age=42)

    users = list(client.users.get_all())
    assert users == [
        User(name="John Doe", age=42),
        User(name="Alice Doe", age=41),
        User(name="Bob Doe", age=10),
        User(name="Carole Doe", age=12),
    ]



from sdkite.http.adapter import HTTPAdapter, HTTPAdapterSpec
from sdkite.http.auth import BasicAuth, NoAuth
from sdkite.http.model import (
    HTTPBodyEncoding,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPResponse,
)

__all__ = (
    # sdkite.http.adapter
    "HTTPAdapter",
    "HTTPAdapterSpec",
    # sdkite.http.auth
    "BasicAuth",
    "NoAuth",
    # sdkite.http.model
    "HTTPBodyEncoding",
    "HTTPHeaderDict",
    "HTTPRequest",
    "HTTPResponse",
)



from functools import wraps
from inspect import Parameter, signature
import sys
from typing import Any, TypeVar, Union, overload

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import Protocol
else:  # pragma: no cover
    from typing import Protocol

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Callable, Iterable, Iterator
else:  # pragma: no cover
    from collections.abc import Callable, Iterable, Iterator

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import Concatenate, ParamSpec
else:  # pragma: no cover
    from typing import Concatenate, ParamSpec


P = ParamSpec("P")
T = TypeVar("T")
U = TypeVar("U")


class Pagination:
    __slots__ = ["_page", "_offset", "_finished", "context"]

    def __init__(
        self,
        *,
        page: int = 0,
        offset: int = 0,
        context: Any = None,
    ) -> None:
        self._page = page
        self._offset = offset
        self._finished = False
        self.context = context

    @property
    def page(self) -> int:
        return self._page

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def finished(self) -> bool:
        return self._finished

    def finish(self) -> None:
        self._finished = True


# typing note: paginated can be applied on either functions or class methods
# for functions, the 'pagination' argument is expected to be the first parameter
# for methods, it is expected to be after the 'self' parameter
# a lot of the type complexity comes from supporting both cases

# note that as a side-effect, it works as well on functions
# where the 'pagination' argument is the second parameter


class _PaginatedDecorator(Protocol):
    @overload
    def __call__(
        self, fct: Callable[Concatenate[Pagination, P], Iterable[T]]
    ) -> Callable[P, Iterator[T]]:
        ...

    @overload
    def __call__(
        self, fct: Callable[Concatenate[U, Pagination, P], Iterable[T]]
    ) -> Callable[Concatenate[U, P], Iterator[T]]:
        ...


def paginated(
    *,
    page: int = 0,
    offset: int = 0,
    context: Any = None,
    stop_when_empty: bool = True,
) -> _PaginatedDecorator:
    @overload
    def paginated_decorator(
        fct: Callable[Concatenate[Pagination, P], Iterable[T]]
    ) -> Callable[P, Iterator[T]]:
        ...

    @overload
    def paginated_decorator(
        fct: Callable[Concatenate[U, Pagination, P], Iterable[T]]
    ) -> Callable[Concatenate[U, P], Iterator[T]]:
        ...

    def paginated_decorator(
        fct: (
            Union[
                Callable[Concatenate[Pagination, P], Iterable[T]],
                Callable[Concatenate[U, Pagination, P], Iterable[T]],
            ]
        )
    ) -> Union[Callable[P, Iterator[T]], Callable[Concatenate[U, P], Iterator[T]]]:
        if isinstance(fct, (classmethod, staticmethod)):
            return type(fct)(paginated_decorator(fct.__func__))  # type: ignore[unreachable]

        sig = signature(fct)
        parameters = tuple(sig.parameters.values())

        # make sure a pagination parameter exists at first or second position

        # while we could accept any position without much code change,
        # it would be more confusing, and we are limited by PEP 677 anyways

        for param_pos in range(2):
            try:
                param = parameters[param_pos]
            except IndexError:
                continue
            if param.name == "pagination" and param.kind in [
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            ]:
                break
        else:
            raise TypeError(
                f"Paginated function '{fct.__name__}' "
                "must have 'pagination' as first parameter"
            )

        # create wrapped function

        @wraps(fct)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
            pagination = Pagination(page=page, offset=offset, context=context)

            try:
                arguments = sig.bind(
                    *args[:param_pos],
                    pagination,
                    *args[param_pos:],
                    **kwargs,
                )
            except TypeError as ex:
                if "pagination" in kwargs:
                    raise TypeError(
                        f"Paginated function '{fct.__name__}' "
                        "cannot be called with a 'pagination' parameter"
                    ) from ex
                raise  # pragma: no cover
            arguments.apply_defaults()

            while not pagination.finished:
                iterable = fct(*arguments.args, **arguments.kwargs)
                empty = True
                for item in iterable:
                    pagination._offset += 1
                    empty = False
                    yield item
                if empty and stop_when_empty:
                    pagination.finish()
                    break
                pagination._page += 1

        # fix the signature and annotations

        # the return value hint is complex to compute exactly from fct's return value
        # e.g. consider 'list[int] | str' should be transformed to 'Iterator[int | str]'
        # so we are just using 'Iterator[Any]'

        # see https://github.com/python/mypy/issues/12472 for ignore below
        wrapped.__signature__ = sig.replace(  # type: ignore
            parameters=parameters[:param_pos] + parameters[param_pos + 1 :],
            return_annotation=Iterator[Any],
        )

        try:
            del wrapped.__annotations__["pagination"]
        except KeyError:
            pass  # maybe no type hint
        wrapped.__annotations__["return"] = Iterator[Any]

        return wrapped

    return paginated_decorator



from sdkite.adapter import Adapter, AdapterSpec
from sdkite.client import Client
from sdkite.pagination import Pagination, paginated

try:
    from sdkite._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.0.dev0-unknown"


__all__ = (
    # sdkite._version
    "__version__",
    # sdkite.adapter
    "Adapter",
    "AdapterSpec",
    # sdkite.client
    "Client",
    # sdkite.pagination
    "Pagination",
    "paginated",
)



from unittest.mock import Mock, call

from sdkite.http import BasicAuth, HTTPHeaderDict, HTTPRequest, NoAuth


def test_basic_auth() -> None:
    class WithBasicAuth:
        http = Mock()
        auth = BasicAuth(http, "user", "password")

    assert WithBasicAuth.http.method_calls == [
        call.register_interceptor("request_interceptor", "auth", 0),
    ]

    client0 = WithBasicAuth()
    client1 = WithBasicAuth()
    adapter = Mock()

    # modify creds on one client instance
    client1.auth.username = "Alice"
    client1.auth.password = "$ecr3t"

    # did not modify creds on class instance
    assert WithBasicAuth.auth.username == "user"
    assert WithBasicAuth.auth.password == "password"

    # did not modify creds on other client instance
    assert client0.auth.username == "user"
    assert client0.auth.password == "password"

    for i, client, cred in [
        (0, client0, "dXNlcjpwYXNzd29yZA=="),
        (1, client1, "QWxpY2U6JGVjcjN0"),
    ]:
        request = HTTPRequest(
            method="GET",
            url="https://www.example.com",
            headers=HTTPHeaderDict({"foo": "bar"}),
            body=b"",
            stream_response=False,
        )
        modified_request = client.auth(request, adapter)
        assert modified_request == HTTPRequest(
            method="GET",
            url="https://www.example.com",
            headers=HTTPHeaderDict(
                {
                    "foo": "bar",
                    "Authorization": f"Basic {cred}",
                }
            ),
            body=b"",
            stream_response=False,
        ), f"client {i}"
        assert not adapter.method_calls, f"client {i}"


def test_basic_auth_encoding() -> None:
    class WithBasicAuth:
        http = Mock()
        auth = BasicAuth(http, "🥸", "🔑")

    client = WithBasicAuth()
    adapter = Mock()

    request = HTTPRequest(
        method="GET",
        url="https://www.example.com",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    modified_request = client.auth(request, adapter)
    assert modified_request == HTTPRequest(
        method="GET",
        url="https://www.example.com",
        headers=HTTPHeaderDict({"Authorization": "Basic 8J+luDrwn5SR"}),
        body=b"",
        stream_response=False,
    )


def test_no_auth() -> None:
    class WithNoAuth:
        http = Mock()
        auth = NoAuth(http)

    assert WithNoAuth.http.method_calls == [
        call.register_interceptor("request_interceptor", "auth", 0),
    ]

    client = WithNoAuth()
    adapter = Mock()

    request = HTTPRequest(
        method="GET",
        url="https://www.example.com",
        headers=HTTPHeaderDict({"foo": "bar"}),
        body=b"",
        stream_response=False,
    )
    modified_request = client.auth(request, adapter)
    assert modified_request == request
    assert not adapter.method_calls



import re
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from sdkite.http import (
    HTTPAdapter,
    HTTPAdapterSpec,
    HTTPBodyEncoding,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPResponse,
)
from sdkite.http import adapter as adapter_module

if TYPE_CHECKING:
    from sdkite import Client
else:
    # we want independent unit tests
    Client = object


class FakeResponse(HTTPResponse):
    def __init__(self, context: str, raw: object) -> None:
        self.context = context
        self._raw = raw

    @property
    def raw(self) -> object:
        return self._raw

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, FakeResponse)
            and self.context == other.context
            and self.raw == other.raw
        )

    def __repr__(self) -> str:
        return f"FakeResponse<{self.context}, {self.raw}>"

    status_code = 200
    reason = "Ok"
    headers = HTTPHeaderDict()
    data_stream = iter(())
    data_bytes = b""
    data_str = ""
    data_json = None


@pytest.fixture(autouse=True)
def patched_adapter(monkeypatch: pytest.MonkeyPatch) -> None:
    impl = Mock()
    impl.return_value = lambda request: FakeResponse("send_request", request)
    monkeypatch.setattr(adapter_module, "RequestsImpl", impl)


def test_no_interceptor() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "send_request",
        HTTPRequest(
            method="GET",
            url="https://www.example.com/xxx/uvw",
            headers=HTTPHeaderDict(),
            body=b"",
            stream_response=False,
        ),
    )


def test_overidden_content_type() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

    client = Klass()
    with pytest.warns(
        RuntimeWarning,
        match=re.escape(
            "The 'content-type' header is being overridden due to request body encoding"
            " HTTPBodyEncoding.JSON (from 'custom' to 'application/json')"
        ),
    ):
        response = client.xxx.request(
            "GET",
            headers={"Content-Type": "custom"},
            body_encoding=HTTPBodyEncoding.JSON,
        )
    assert response == FakeResponse(
        "send_request",
        HTTPRequest(
            method="GET",
            url="https://www.example.com/xxx",
            headers=HTTPHeaderDict({"content-type": "application/json"}),
            body=b"null",
            stream_response=False,
        ),
    )


def test_interceptor_method() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        def xxx_req(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        def xxx_resp(self, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "xxx_resp",
        FakeResponse(
            "send_request",
            HTTPRequest(
                method="GET",
                url="https://www.example.com/xxx/uvw",
                headers=HTTPHeaderDict({"intercept": "xxx"}),
                body=b"",
                stream_response=False,
            ),
        ),
    )


def test_interceptor_staticmethod() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")
        yyy = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        @staticmethod
        def xxx_req(request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        @staticmethod
        def xxx_resp(response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

        @staticmethod
        @yyy.intercept_request(0)
        def yyy_intercept(request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "yyy")
            return request

        @staticmethod
        @yyy.intercept_response(1)
        def yyy_resp(response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("yyy_resp", response)

    client = Klass()
    for attr in ("xxx", "yyy"):
        response = getattr(client, attr).request("GET", "uvw")
        assert response == FakeResponse(
            f"{attr}_resp",
            FakeResponse(
                "send_request",
                HTTPRequest(
                    method="GET",
                    url="https://www.example.com/xxx/uvw",
                    headers=HTTPHeaderDict({"intercept": attr}),
                    body=b"",
                    stream_response=False,
                ),
            ),
        )


def test_interceptor_classmethod() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")
        yyy = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        @classmethod
        def xxx_req(cls, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        @classmethod
        def xxx_resp(cls, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

        @classmethod
        @yyy.intercept_request(0)
        def yyy_intercept(cls, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "yyy")
            return request

        @classmethod
        @yyy.intercept_response(1)
        def yyy_resp(cls, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("yyy_resp", response)

    client = Klass()
    for attr in ("xxx", "yyy"):
        response = getattr(client, attr).request("GET", "uvw")
        assert response == FakeResponse(
            f"{attr}_resp",
            FakeResponse(
                "send_request",
                HTTPRequest(
                    method="GET",
                    url="https://www.example.com/xxx/uvw",
                    headers=HTTPHeaderDict({"intercept": attr}),
                    body=b"",
                    stream_response=False,
                ),
            ),
        )


def test_register_interceptor_existing() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        def xxx_req(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        def xxx_resp(self, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

    with pytest.warns(
        RuntimeWarning,
        match=re.escape(
            "Interceptor 'xxx_req' of 'xxx' has already been registered"
            " with order 0, ignoring new registration with order 42"
        ),
    ):
        Klass.xxx.register_interceptor("request_interceptor", "xxx_req", 42)

    with pytest.warns(
        RuntimeWarning,
        match=re.escape(
            "Interceptor 'xxx_resp' of 'xxx' has already been registered"
            " with order 1, ignoring new registration with order 42"
        ),
    ):
        Klass.xxx.register_interceptor("response_interceptor", "xxx_resp", 42)

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "xxx_resp",
        FakeResponse(
            "send_request",
            HTTPRequest(
                method="GET",
                url="https://www.example.com/xxx/uvw",
                headers=HTTPHeaderDict({"intercept": "xxx"}),  # only once
                body=b"",
                stream_response=False,
            ),
        ),
    )



import sys
from typing import Tuple, TypeVar

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Sequence
else:  # pragma: no cover
    from collections.abc import Iterable, Sequence


T = TypeVar("T")
U = TypeVar("U")


def zip_reverse(items_a: Sequence[T], items_b: Sequence[U]) -> Iterable[Tuple[T, U]]:
    if len(items_a) != len(items_b):
        # in Python >= 3.10 we could use zip(..., strict=True)
        # but since we have sequences anyways this avoids dealing with backward compatibility
        raise ValueError("zip_reverse() arguments have different lengths")
    return zip(reversed(items_a), reversed(items_b))



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto, unique
import sys
from typing import Dict, List, Tuple, Union

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator, Mapping, MutableMapping
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator, Mapping, MutableMapping


class HTTPHeaderDict(MutableMapping[str, str]):
    __slots__ = ["_contents"]

    def __init__(
        self,
        items: Union[None, Mapping[str, str], Iterable[Tuple[str, str]]] = None,
    ) -> None:
        self._contents: Dict[str, List[str]] = {}
        if isinstance(items, Mapping):
            items = items.items()
        if items:
            for key, value in items:
                self.add(key, value)

    def __repr__(self) -> str:
        as_dict = {values[0]: values[1:] for values in self._contents.values()}
        return f"HTTPHeaderDict{as_dict!r}"

    def __iter__(self) -> Iterator[str]:
        for values in self._contents.values():
            yield values[0]

    def __len__(self) -> int:
        return len(self._contents)

    def __getitem__(self, key: str) -> str:
        try:
            return ", ".join(self._contents[key.lower()][1:])
        except KeyError:
            raise KeyError(key)  # pylint: disable=raise-missing-from

    def __setitem__(self, key: str, value: str) -> None:
        self._contents[key.lower()] = [key, value]

    def __delitem__(self, key: str) -> None:
        del self._contents[key.lower()]

    def add(self, key: str, value: str) -> None:
        try:
            self._contents[key.lower()].append(value)
        except KeyError:
            self[key] = value


@unique
class HTTPBodyEncoding(Enum):
    AUTO = auto()
    NONE = auto()
    JSON = auto()
    URLENCODE = auto()
    MULTIPART = auto()


@dataclass
class HTTPRequest:
    method: str
    url: str
    headers: HTTPHeaderDict
    body: Union[bytes, Iterator[bytes]]
    stream_response: bool


class HTTPResponse(ABC):
    @property
    @abstractmethod
    def raw(self) -> object:
        """
        The response object coming from the adapter
        """

    @property
    @abstractmethod
    def status_code(self) -> int:
        """
        The HTTP status code (RFC 2616 section 6.1.1)
        """

    @property
    @abstractmethod
    def reason(self) -> str:
        """
        The HTTP reason phrase (RFC 2616 section 6.1.1)
        """

    @property
    @abstractmethod
    def headers(self) -> HTTPHeaderDict:
        """
        The HTTP headers as a HTTPHeaderDict instance
        """

    @property
    @abstractmethod
    def data_stream(self) -> Iterator[bytes]:
        """
        The body of the response as an iterator of bytes, useful for data streaming.
        """

    @property
    @abstractmethod
    def data_bytes(self) -> bytes:
        """
        The body of the response as bytes.
        """

    @property
    @abstractmethod
    def data_str(self) -> str:
        """
        The body of the response as a str, for easier access.
        """

    @property
    @abstractmethod
    def data_json(self) -> object:
        """
        The body of the response JSON-decoded, for easier access.
        """
